package dev.shadowsoffire.apotheosis.adventure.loot;

import java.util.Collections;
import dev.shadowsoffire.apotheosis.adventure.AdventureModule;
import org.apache.commons.lang3.mutable.MutableInt;
import dev.shadowsoffire.apotheosis.adventure.affix.Affix;
import java.util.Set;
import net.minecraft.world.item.ItemStack;
import java.util.Optional;
import java.util.Random;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixType;
import net.minecraft.util.ExtraCodecs;
import com.mojang.datafixers.kinds.Applicative;
import com.mojang.serialization.codecs.ListCodec;
import dev.shadowsoffire.placebo.codec.PlaceboCodecs;
import net.minecraftforge.registries.ForgeRegistries;
import com.mojang.datafixers.kinds.App;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import net.minecraft.resources.ResourceLocation;
import java.util.function.Function;
import java.util.Objects;
import java.util.Map;
import net.minecraft.util.RandomSource;
import net.minecraft.network.chat.Style;
import net.minecraft.network.chat.Component;
import javax.annotation.Nullable;
import java.util.List;
import net.minecraft.network.chat.TextColor;
import net.minecraft.world.item.Item;
import dev.shadowsoffire.placebo.reload.DynamicHolder;
import com.mojang.serialization.Codec;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;
import dev.shadowsoffire.placebo.codec.CodecProvider;

public class LootRarity implements CodecProvider<LootRarity>, WeightedDynamicRegistry.ILuckyWeighted, Comparable<LootRarity>
{
    public static final Codec<LootRarity> LOAD_CODEC;
    @Deprecated
    public static final Codec<DynamicHolder<LootRarity>> HOLDER_CODEC;
    public static final Codec<LootRarity> CODEC;
    private final Item material;
    private final TextColor color;
    private final int ordinal;
    private final int weight;
    private final float quality;
    private final List<LootRule> rules;
    
    private LootRarity(final TextColor color, final Item material, final int ordinal, final int weight, final float quality, final List<LootRule> rules) {
        this.color = color;
        this.material = material;
        this.ordinal = ordinal;
        this.weight = weight;
        this.quality = quality;
        this.rules = rules;
    }
    
    public Item getMaterial() {
        return this.material;
    }
    
    public TextColor getColor() {
        return this.color;
    }
    
    public int ordinal() {
        return this.ordinal;
    }
    
    public int getWeight() {
        return this.weight;
    }
    
    public float getQuality() {
        return this.quality;
    }
    
    public List<LootRule> getRules() {
        return this.rules;
    }
    
    public LootRarity next() {
        return (LootRarity)RarityRegistry.next((DynamicHolder<LootRarity>)RarityRegistry.INSTANCE.holder((CodecProvider)this)).get();
    }
    
    public LootRarity prev() {
        return (LootRarity)RarityRegistry.prev((DynamicHolder<LootRarity>)RarityRegistry.INSTANCE.holder((CodecProvider)this)).get();
    }
    
    public boolean isAtMost(final LootRarity other) {
        return this.ordinal() <= other.ordinal();
    }
    
    public boolean isAtLeast(final LootRarity other) {
        return this.ordinal() >= other.ordinal();
    }
    
    public static LootRarity min(final LootRarity a, @Nullable final LootRarity b) {
        if (b == null) {
            return a;
        }
        return (a.ordinal <= b.ordinal) ? a : b;
    }
    
    public static LootRarity max(final LootRarity a, @Nullable final LootRarity b) {
        if (b == null) {
            return a;
        }
        return (a.ordinal >= b.ordinal) ? a : b;
    }
    
    public LootRarity clamp(@Nullable final LootRarity lowerBound, @Nullable final LootRarity upperBound) {
        return max(min(this, upperBound), lowerBound);
    }
    
    public Component toComponent() {
        return (Component)Component.m_237115_("rarity." + RarityRegistry.INSTANCE.getKey((CodecProvider)this)).m_130948_(Style.f_131099_.m_131148_(this.color));
    }
    
    @Override
    public String toString() {
        return "LootRarity{" + RarityRegistry.INSTANCE.getKey((CodecProvider)this);
    }
    
    public int compareTo(final LootRarity o) {
        return Integer.compare(this.ordinal, o.ordinal);
    }
    
    public Codec<? extends LootRarity> getCodec() {
        return LootRarity.LOAD_CODEC;
    }
    
    public static LootRarity random(final RandomSource rand, final float luck) {
        return (LootRarity)RarityRegistry.INSTANCE.getRandomItem(rand, luck);
    }
    
    public static LootRarity random(final RandomSource rand, final float luck, @Nullable final RarityClamp clamp) {
        final LootRarity rarity = random(rand, luck);
        return (clamp == null) ? rarity : clamp.clamp(rarity);
    }
    
    public static <T> Codec<Map<LootRarity, T>> mapCodec(final Codec<T> codec) {
        return (Codec<Map<LootRarity, T>>)Codec.unboundedMap((Codec)LootRarity.CODEC, (Codec)codec);
    }
    
    static {
        LOAD_CODEC = RecordCodecBuilder.create(inst -> inst.group((App)TextColor.f_237295_.fieldOf("color").forGetter((Function)LootRarity::getColor), (App)ForgeRegistries.ITEMS.getCodec().fieldOf("material").forGetter((Function)LootRarity::getMaterial), (App)Codec.INT.fieldOf("ordinal").forGetter((Function)LootRarity::ordinal), (App)Codec.intRange(0, Integer.MAX_VALUE).fieldOf("weight").forGetter((Function)WeightedDynamicRegistry.ILuckyWeighted::getWeight), (App)PlaceboCodecs.nullableField(Codec.floatRange(0.0f, Float.MAX_VALUE), "quality", (Object)0.0f).forGetter((Function)WeightedDynamicRegistry.ILuckyWeighted::getQuality), (App)new ListCodec((Codec)LootRule.CODEC).fieldOf("rules").forGetter((Function)LootRarity::getRules)).apply((Applicative)inst, LootRarity::new));
        HOLDER_CODEC = ExtraCodecs.m_184415_(() -> {
            Codec.STRING.xmap((Function)RarityRegistry::convertId, (Function)ResourceLocation::toString);
            final RarityRegistry instance = RarityRegistry.INSTANCE;
            Objects.requireNonNull(instance);
            final Codec codec;
            return codec.xmap((Function)instance::holder, (Function)DynamicHolder::getId);
        });
        CODEC = ExtraCodecs.m_184415_(() -> {
            final Codec<DynamicHolder<LootRarity>> holder_CODEC = LootRarity.HOLDER_CODEC;
            DynamicHolder::get;
            final RarityRegistry instance2 = RarityRegistry.INSTANCE;
            Objects.requireNonNull(instance2);
            final Function function;
            return holder_CODEC.xmap(function, (Function)instance2::holder);
        });
    }
    
    record LootRule(AffixType type, float chance, @Nullable LootRule backup) {
        public static final Codec<LootRule> CODEC;
        private static Random jRand;
        
        public LootRule(final AffixType type, final float chance) {
            this(type, chance, Optional.empty());
        }
        
        public LootRule(final AffixType type, final float chance, final Optional<LootRule> backup) {
            this(type, chance, backup.orElse(null));
        }
        
        public void execute(final ItemStack stack, final LootRarity rarity, final Set<DynamicHolder<? extends Affix>> currentAffixes, final MutableInt sockets, final RandomSource rand) {
            if (this.type == AffixType.DURABILITY) {
                return;
            }
            if (rand.m_188501_() <= this.chance) {
                if (this.type == AffixType.SOCKET) {
                    sockets.add(1);
                    return;
                }
                final List<DynamicHolder<? extends Affix>> available = LootController.getAvailableAffixes(stack, rarity, currentAffixes, this.type);
                if (available.size() == 0) {
                    if (this.backup != null) {
                        this.backup.execute(stack, rarity, currentAffixes, sockets, rand);
                    }
                    else {
                        AdventureModule.LOGGER.error("Failed to execute LootRule {}/{}/{}/{}!", (Object)ForgeRegistries.ITEMS.getKey((Object)stack.m_41720_()), (Object)RarityRegistry.INSTANCE.getKey((CodecProvider)rarity), (Object)this.type, (Object)this.chance);
                    }
                    return;
                }
                LootRule.jRand.setSeed(rand.m_188505_());
                Collections.shuffle(available, LootRule.jRand);
                currentAffixes.add(available.get(0));
            }
        }
        
        @Nullable
        public LootRule backup() {
            return this.backup;
        }
        
        static {
            CODEC = RecordCodecBuilder.create(inst -> inst.group((App)PlaceboCodecs.enumCodec((Class)AffixType.class).fieldOf("type").forGetter((Function)LootRule::type), (App)Codec.FLOAT.fieldOf("chance").forGetter((Function)LootRule::chance), (App)PlaceboCodecs.nullableField(ExtraCodecs.m_184415_(() -> LootRule.CODEC), "backup").forGetter(rule -> Optional.ofNullable(rule.backup()))).apply((Applicative)inst, LootRule::new));
            LootRule.jRand = new Random();
        }
    }
}
